/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.apache.yoko.bindings.corba;

import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.math.BigInteger;
import java.util.Iterator;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

//import org.apache.schemas.yoko.bindings.corba.Exception;
import org.apache.yoko.bindings.corba.types.CorbaExceptionHandler;
import org.apache.yoko.bindings.corba.types.CorbaFixedHandler;
import org.apache.yoko.bindings.corba.types.CorbaObjectHandler;
import org.apache.yoko.bindings.corba.types.CorbaPrimitiveHandler;

import org.objectweb.celtix.common.logging.LogUtils;

import org.omg.CORBA.TCKind;

// This class exists to allow for the conversion of a Corba UserException into an exception that a 
// Yoko client/server can use.  The big issue with this conversion occurs when going from a Corba 
// UserException to a Yoko exception.  Currently, the data readers/writers that we use to marshal 
// and unmarshal regular data will not support the marshalling/unmarshalling of exceptions when 
// we use XML Events.  This class allows us to unmarshal the data correctly into an appropriate Yoko
// exception object.

public class CorbaExceptionHelper {

    private static final Logger LOG = LogUtils.getL7dLogger(CorbaExceptionHelper.class);
    
    public static void unmarshalCorbaException(CorbaExceptionHandler corbaEx, Object yokoEx)
        throws CorbaBindingException {
        try {
            LOG.log(Level.INFO, "Preparing to unmarshal a CORBA exception");

            List<CorbaObjectHandler> exMembers = corbaEx.getMembers();
            for (Iterator<CorbaObjectHandler> iter = exMembers.iterator(); iter.hasNext();) {
                CorbaObjectHandler member = iter.next();
                String memberName = member.getName().getLocalPart();
                LOG.log(Level.INFO, "Setting field " + memberName);

                Field field = yokoEx.getClass().getDeclaredField(memberName);
                Method setMethod = getFieldSetMethod(memberName, yokoEx.getClass(), field);
                setExceptionFieldValue(member, field, setMethod, yokoEx);
            }
        } catch (java.lang.Exception ex) {
            LOG.log(Level.SEVERE, "Error during CORBA unmarshal exception: " + ex.getMessage());
            throw new CorbaBindingException("Unable to unmarshal CORBA exception: " + ex.getMessage(), ex);
        }
    }   

    private static void setExceptionFieldValue(CorbaObjectHandler obj, Field field, 
                                               Method setMethod, Object objInstance) 
        throws CorbaBindingException {
        // There are special cases where the wrapper for the primitive type doesn't match the 
        // type used by the wsdl2java tool.  Make accomodations for these.  Otherwise, just 
        // invoke the setMethod with the object stored in the holder.  REVISIT: we may want 
        // to change the way the binding handles the data internally to make this step 
        // unnecessary.
        try {
            switch (obj.getTypeCode().kind().value()) {
            case TCKind._tk_boolean:
                Object boolValue = ((CorbaPrimitiveHandler)obj).getValue();
                setMethod.invoke(objInstance, boolValue);
                break;
            case TCKind._tk_char:
                byte charValue = (byte)((Character)((CorbaPrimitiveHandler)obj).getValue()).charValue();
                setMethod.invoke(objInstance, Byte.valueOf(charValue));
                break;
            case TCKind._tk_wchar:
                String wcharValue = ((Character)((CorbaPrimitiveHandler)obj).getValue()).toString();
                setMethod.invoke(objInstance, wcharValue);
                break;
            case TCKind._tk_octet:
                short octetValue = Short.valueOf((Byte)((CorbaPrimitiveHandler)obj).getValue()).shortValue();
                setMethod.invoke(objInstance, octetValue);
                break;
            case TCKind._tk_short:
            case TCKind._tk_ushort:
                Object shortValue = ((CorbaPrimitiveHandler)obj).getValue();
                setMethod.invoke(objInstance, shortValue);
                break;
            case TCKind._tk_long:
                int longValue = ((Long)((CorbaPrimitiveHandler)obj).getValue()).intValue();
                setMethod.invoke(objInstance, Integer.valueOf(longValue));
                break;
            case TCKind._tk_ulong:
            case TCKind._tk_longlong:
                long ulongValue = ((BigInteger)((CorbaPrimitiveHandler)obj).getValue()).longValue();
                setMethod.invoke(objInstance, Long.valueOf(ulongValue));
                break;
            case TCKind._tk_ulonglong:
                Object ulonglongValue = ((CorbaPrimitiveHandler)obj).getValue();
                setMethod.invoke(objInstance, ulonglongValue);
                break;
            case TCKind._tk_float:
                Object floatValue = ((CorbaPrimitiveHandler)obj).getValue();
                setMethod.invoke(objInstance, floatValue);
                break;
            case TCKind._tk_double:
                Object doubleValue = ((CorbaPrimitiveHandler)obj).getValue();
                setMethod.invoke(objInstance, doubleValue);
                break;
            case TCKind._tk_string:
            case TCKind._tk_wstring:
                Object stringValue = ((CorbaPrimitiveHandler)obj).getValue();
                setMethod.invoke(objInstance, stringValue);
                break;
            case TCKind._tk_fixed:
                CorbaFixedHandler fixedHandler = (CorbaFixedHandler)obj;
                setMethod.invoke(objInstance, fixedHandler.getValue());
                break;
            default:
                setMethod.invoke(objInstance, createExceptionField(obj, field.getClass()));
                break;
            }
        } catch (java.lang.Exception ex) {
            throw new CorbaBindingException("Unable to set the exception field " 
                                            + obj.getName().getLocalPart(), ex);
        }
    }
    
    private static Object createExceptionField(CorbaObjectHandler obj, Class cls) 
        throws CorbaBindingException {
        // This method will allow for the creation of nested complex types within the exception.
        // TODO: Provide implementation
        throw new CorbaBindingException("Non-primitive exception fields not implemented yet.");
    }

    private static Method getFieldSetMethod(String fieldName, Class cls, Field field) 
        throws CorbaBindingException {
        Method setMethod = null;
        String methodName = "set" + (char)(fieldName.charAt(0) - 'a' + 'A') + fieldName.substring(1);
        try {
            setMethod = cls.getMethod(methodName, field.getType());
        } catch (java.lang.Exception ex) {
            throw new CorbaBindingException("Unable to locate the 'set' method " + methodName, ex);
        }

        return setMethod;
    }
}
