/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */
package org.apache.yoko.bindings.corba;

import javax.xml.ws.Holder;

import junit.framework.TestCase;

import org.apache.yoko.wsdl.CorbaConstants;

import org.objectweb.celtix.Bus;
import org.objectweb.celtix.bindings.BindingManager;
import org.objectweb.celtix.bindings.DataBindingCallback;
import org.objectweb.celtix.bus.jaxws.JAXBDataBindingCallback;
import org.objectweb.celtix.context.GenericMessageContext;
import org.objectweb.celtix.context.ObjectMessageContext;
import org.objectweb.celtix.context.ObjectMessageContextImpl;
import org.objectweb.celtix.ws.addressing.EndpointReferenceType;

import org.omg.CORBA.ARG_IN;
import org.omg.CORBA.ARG_INOUT;
import org.omg.CORBA.ARG_OUT;
import org.omg.CORBA.ORB;
import org.omg.CORBA.TCKind;

public class CorbaBindingImplMarshalTest extends TestCase {

    private static String TEST_CLASS = "org.apache.schemas.yoko.idl.parammodes.ParamModes";
    
    private TestUtils testUtils;
    private ORB orb;
    private Bus bus;
    private EndpointReferenceType endpoint;
    private CorbaBindingImpl bindingImpl;
    private Class paramModesClass;
    
    public CorbaBindingImplMarshalTest(String arg0) {
        super(arg0);
    }
    public static void main(String[] args) {
        junit.textui.TestRunner.run(CorbaBindingImplMarshalTest.class);
    }
    
    protected void setUp() throws Exception {
        super.setUp();
    
        testUtils = new TestUtils();
    
        java.util.Properties props = System.getProperties();
        props.put("org.omg.CORBA.ORBClass", "org.apache.yoko.orb.CORBA.ORB");
        props.put("org.omg.CORBA.ORBSingletonClass", "org.apache.yoko.orb.CORBA.ORBSingleton");
        props.put("yoko.orb.id", "Yoko-Test");
        orb = ORB.init(new String[0], props);
        bus = Bus.init(new String[0]);
        
        paramModesClass = Thread.currentThread().getContextClassLoader().loadClass(TEST_CLASS);
        
        // programmatically add the corba binding factory
        BindingManager manager = bus.getBindingManager();
        CorbaBindingFactory corbaBF = new CorbaBindingFactory();
        corbaBF.init(bus);
        manager.registerBinding(CorbaConstants.NU_WSDL_CORBA, corbaBF);
        
        endpoint = testUtils.getParamModesEndpointReference();
        bindingImpl = new CorbaBindingImpl(bus, endpoint, orb, false);
    }
    
    protected void tearDown() throws Exception {
        if (orb != null) {
            try {
                orb.destroy();
            } catch (Exception ex) {
                // Do nothing.  Throw an Exception?
            }
        }
    }
    
    public void testMarshalSingleInParam() throws Exception {
        ObjectMessageContextImpl objContext = null;
        CorbaMessageContextImpl corbaContext = null;
        Class[] params = null;
        
        params = new Class[1];
        params[0] = short.class;
        
        // First try call the operation using a void return type
        objContext = new ObjectMessageContextImpl();
        objContext.setMethod(paramModesClass.getMethod("testSingleInParam", params));
        objContext.setMessageObjects((short)123);
        
        corbaContext = new CorbaMessageContextImpl(new GenericMessageContext());
        corbaContext.put(ObjectMessageContext.MESSAGE_INPUT, true);
        
        bindingImpl.marshal(objContext,
                            corbaContext,
                            new JAXBDataBindingCallback(objContext.getMethod(),
                                                        DataBindingCallback.Mode.MESSAGE,
                                                        null));
        
        // Verify that everything was marshalled correctly
        CorbaStreamable[] streamables = corbaContext.getMessage().getStreamableArguments();
        assertTrue(streamables.length == 1);
        CorbaStreamable param0 = streamables[0];
        assertTrue(param0.getObject().getTypeCodeKind().value() == TCKind._tk_short);
        assertTrue(param0.getMode() == ARG_IN.value);

        // Now try the call with a return value
        objContext = new ObjectMessageContextImpl();
        objContext.setMethod(paramModesClass.getMethod("testSingleInParamWithReturn", params));
        objContext.setMessageObjects((short)123);
        
        corbaContext = new CorbaMessageContextImpl(new GenericMessageContext());
        corbaContext.put(ObjectMessageContext.MESSAGE_INPUT, true);
        
        bindingImpl.marshal(objContext,
                            corbaContext,
                            new JAXBDataBindingCallback(objContext.getMethod(),
                                                        DataBindingCallback.Mode.MESSAGE,
                                                        null));
        
        // Verify that everything was marshalled correctly
        streamables = corbaContext.getMessage().getStreamableArguments();
        assertTrue(streamables.length == 1);
        param0 = streamables[0];
        assertTrue(param0.getObject().getTypeCodeKind().value() == TCKind._tk_short);
        assertTrue(param0.getMode() == ARG_IN.value);
        CorbaStreamable ret = corbaContext.getMessage().getStreamableReturn();
        assertNotNull(ret);
        assertTrue(ret.getObject().getTypeCodeKind().value() == TCKind._tk_short);
    }
    
    public void testMarshalSingleOutParam() throws Exception {
        ObjectMessageContextImpl objContext = null;
        CorbaMessageContextImpl corbaContext = null;
        Class[] params = null;
        
        // First try call the operation using a void return type
        // Since this call has a void return type, the outparam is actually mapped to the return type
        params = new Class[0];
        
        objContext = new ObjectMessageContextImpl();
        objContext.setMethod(paramModesClass.getMethod("testSingleOutParam", params));
        
        corbaContext = new CorbaMessageContextImpl(new GenericMessageContext());
        corbaContext.put(ObjectMessageContext.MESSAGE_INPUT, true);
        
        bindingImpl.marshal(objContext,
                            corbaContext,
                            new JAXBDataBindingCallback(objContext.getMethod(),
                                                        DataBindingCallback.Mode.MESSAGE,
                                                        null));
        
        // Verify that everything was marshalled correctly
        CorbaStreamable[] streamables = corbaContext.getMessage().getStreamableArguments();
        assertTrue(streamables.length == 1);
        CorbaStreamable param0 = streamables[0];
        assertTrue(param0.getObject().getTypeCodeKind().value() == TCKind._tk_short);
        assertTrue(param0.getMode() == ARG_OUT.value);

        // Now try the call with a return value
        params = new Class[1];
        params[0] = new Holder<Short>().getClass();
        
        objContext = new ObjectMessageContextImpl();
        objContext.setMethod(paramModesClass.getMethod("testSingleOutParamWithReturn", params));
        objContext.setMessageObjects(new Holder<Short>((short)123));
        corbaContext = new CorbaMessageContextImpl(new GenericMessageContext());
        corbaContext.put(ObjectMessageContext.MESSAGE_INPUT, true);
        
        bindingImpl.marshal(objContext,
                            corbaContext,
                            new JAXBDataBindingCallback(objContext.getMethod(),
                                                        DataBindingCallback.Mode.MESSAGE,
                                                        null));
        
        // Verify that everything was marshalled correctly
        streamables = corbaContext.getMessage().getStreamableArguments();
        assertTrue(streamables.length == 1);
        param0 = streamables[0];
        assertTrue(param0.getObject().getTypeCodeKind().value() == TCKind._tk_short);
        assertTrue(param0.getMode() == ARG_OUT.value);
        CorbaStreamable ret = corbaContext.getMessage().getStreamableReturn();
        assertNotNull(ret);
        assertTrue(ret.getObject().getTypeCodeKind().value() == TCKind._tk_short);
    }

    public void testMarshalSingleInOutParam() throws Exception {
        ObjectMessageContextImpl objContext = null;
        CorbaMessageContextImpl corbaContext = null;
        Class[] params = null;
        
        // First try call the operation using a void return type
        params = new Class[1];
        params[0] = new Holder<Short>().getClass();
        
        objContext = new ObjectMessageContextImpl();
        objContext.setMethod(paramModesClass.getMethod("testSingleInOutParam", params));
        objContext.setMessageObjects(new Holder<Short>(new Short((short)123)));
        
        corbaContext = new CorbaMessageContextImpl(new GenericMessageContext());
        corbaContext.put(ObjectMessageContext.MESSAGE_INPUT, true);
        
        bindingImpl.marshal(objContext,
                            corbaContext,
                            new JAXBDataBindingCallback(objContext.getMethod(),
                                                        DataBindingCallback.Mode.MESSAGE,
                                                        null));
        
        // Verify that everything was marshalled correctly
        CorbaStreamable[] streamables = corbaContext.getMessage().getStreamableArguments();
        assertTrue(streamables.length == 1);
        CorbaStreamable param0 = streamables[0];
        assertTrue(param0.getObject().getTypeCodeKind().value() == TCKind._tk_short);
        assertTrue(param0.getMode() == ARG_INOUT.value);
        
        // Now try the call with the return value
        objContext = new ObjectMessageContextImpl();
        objContext.setMethod(paramModesClass.getMethod("testSingleInOutParamWithReturn", params));
        objContext.setMessageObjects(new Holder<Short>(new Short((short)123)));
        
        corbaContext = new CorbaMessageContextImpl(new GenericMessageContext());
        corbaContext.put(ObjectMessageContext.MESSAGE_INPUT, true);
        
        bindingImpl.marshal(objContext,
                            corbaContext,
                            new JAXBDataBindingCallback(objContext.getMethod(),
                                                        DataBindingCallback.Mode.MESSAGE,
                                                        null));
        
        // Verify that everything was marshalled correctly
        streamables = corbaContext.getMessage().getStreamableArguments();
        assertTrue(streamables.length == 1);
        param0 = streamables[0];
        assertTrue(param0.getObject().getTypeCodeKind().value() == TCKind._tk_short);
        assertTrue(param0.getMode() == ARG_INOUT.value);
        CorbaStreamable ret = corbaContext.getMessage().getStreamableReturn();
        assertNotNull(ret);
        assertTrue(ret.getObject().getTypeCodeKind().value() == TCKind._tk_short);
    }

    public void testMarshalMultipleInParams() throws Exception {
        ObjectMessageContextImpl objContext = null;
        CorbaMessageContextImpl corbaContext = null;
        Class[] params = null;
        
        // First try call the operation using a void return type
        params = new Class[2];
        params[0] = short.class;
        params[1] = short.class;
        
        objContext = new ObjectMessageContextImpl();
        objContext.setMethod(paramModesClass.getMethod("testMultipleInParams", params));
        Object[] paramObjs = new Object[2];
        paramObjs[0] = (short)123;
        paramObjs[1] = (short)456;
        objContext.setMessageObjects(paramObjs);
        
        corbaContext = new CorbaMessageContextImpl(new GenericMessageContext());
        corbaContext.put(ObjectMessageContext.MESSAGE_INPUT, true);
        
        bindingImpl.marshal(objContext,
                            corbaContext,
                            new JAXBDataBindingCallback(objContext.getMethod(),
                                                        DataBindingCallback.Mode.MESSAGE,
                                                        null));
        
        // Verify that everything was marshalled correctly
        CorbaStreamable[] streamables = corbaContext.getMessage().getStreamableArguments();
        assertTrue(streamables.length == 2);
        CorbaStreamable param0 = streamables[0];
        assertTrue(param0.getObject().getTypeCodeKind().value() == TCKind._tk_short);
        assertTrue(param0.getMode() == ARG_IN.value);
        CorbaStreamable param1 = streamables[1];
        assertTrue(param1.getObject().getTypeCodeKind().value() == TCKind._tk_short);
        assertTrue(param1.getMode() == ARG_IN.value);
        
        // Now try the call with a return value
        objContext = new ObjectMessageContextImpl();
        objContext.setMethod(paramModesClass.getMethod("testMultipleInParamsWithReturn", params));
        objContext.setMessageObjects(paramObjs);
        
        corbaContext = new CorbaMessageContextImpl(new GenericMessageContext());
        corbaContext.put(ObjectMessageContext.MESSAGE_INPUT, true);
        
        bindingImpl.marshal(objContext,
                            corbaContext,
                            new JAXBDataBindingCallback(objContext.getMethod(),
                                                        DataBindingCallback.Mode.MESSAGE,
                                                        null));
        
        // Verify that everything was marshalled correctly
        streamables = corbaContext.getMessage().getStreamableArguments();
        assertTrue(streamables.length == 2);
        param0 = streamables[0];
        assertTrue(param0.getObject().getTypeCodeKind().value() == TCKind._tk_short);
        assertTrue(param0.getMode() == ARG_IN.value);
        param1 = streamables[1];
        assertTrue(param1.getObject().getTypeCodeKind().value() == TCKind._tk_short);
        assertTrue(param1.getMode() == ARG_IN.value);
        CorbaStreamable ret = corbaContext.getMessage().getStreamableReturn();
        assertNotNull(ret);
        assertTrue(ret.getObject().getTypeCodeKind().value() == TCKind._tk_short);
    }

    public void testMarshalMultipleOutParams() throws Exception {
        ObjectMessageContextImpl objContext = null;
        CorbaMessageContextImpl corbaContext = null;
        Class[] params = null;
        
        // First try call the operation using a void return type
        // Since the call has more than one out parameter, they should be mapped to the parameters
        params = new Class[2];
        params[0] = new Holder<Short>().getClass();
        params[1] = new Holder<Short>().getClass();
        
        objContext = new ObjectMessageContextImpl();
        objContext.setMethod(paramModesClass.getMethod("testMultipleOutParams", params));
        Object[] paramObjs = new Object[2];
        paramObjs[0] = new Holder<Short>((short)123);
        paramObjs[1] = new Holder<Short>((short)456);
        objContext.setMessageObjects(paramObjs);
        
        corbaContext = new CorbaMessageContextImpl(new GenericMessageContext());
        corbaContext.put(ObjectMessageContext.MESSAGE_INPUT, true);
        
        bindingImpl.marshal(objContext,
                            corbaContext,
                            new JAXBDataBindingCallback(objContext.getMethod(),
                                                        DataBindingCallback.Mode.MESSAGE,
                                                        null));
        
        // Verify that everything was marshalled correctly
        CorbaStreamable[] streamables = corbaContext.getMessage().getStreamableArguments();
        assertTrue(streamables.length == 2);
        CorbaStreamable param0 = streamables[0];
        assertTrue(param0.getObject().getTypeCodeKind().value() == TCKind._tk_short);
        assertTrue(param0.getMode() == ARG_OUT.value);
        CorbaStreamable param1 = streamables[1];
        assertTrue(param1.getObject().getTypeCodeKind().value() == TCKind._tk_short);
        assertTrue(param1.getMode() == ARG_OUT.value);

        // Now try the call with a return value
        objContext = new ObjectMessageContextImpl();
        objContext.setMethod(paramModesClass.getMethod("testMultipleOutParamsWithReturn", params));
        objContext.setMessageObjects(paramObjs);
        
        corbaContext = new CorbaMessageContextImpl(new GenericMessageContext());
        corbaContext.put(ObjectMessageContext.MESSAGE_INPUT, true);
        
        bindingImpl.marshal(objContext,
                            corbaContext,
                            new JAXBDataBindingCallback(objContext.getMethod(),
                                                        DataBindingCallback.Mode.MESSAGE,
                                                        null));
        
        // Verify that everything was marshalled correctly
        streamables = corbaContext.getMessage().getStreamableArguments();
        assertTrue(streamables.length == 2);
        param0 = streamables[0];
        assertTrue(param0.getObject().getTypeCodeKind().value() == TCKind._tk_short);
        assertTrue(param0.getMode() == ARG_OUT.value);
        param1 = streamables[1];
        assertTrue(param1.getObject().getTypeCodeKind().value() == TCKind._tk_short);
        assertTrue(param1.getMode() == ARG_OUT.value);
        CorbaStreamable ret = corbaContext.getMessage().getStreamableReturn();
        assertNotNull(ret);
        assertTrue(ret.getObject().getTypeCodeKind().value() == TCKind._tk_short);
    }

    public void testMarshalMultipleInOutParams() throws Exception {
        ObjectMessageContextImpl objContext = null;
        CorbaMessageContextImpl corbaContext = null;
        Class[] params = null;
        
        // First try call the operation using a void return type
        params = new Class[2];
        params[0] = new Holder<Short>().getClass();
        params[1] = new Holder<Short>().getClass();
        
        objContext = new ObjectMessageContextImpl();
        objContext.setMethod(paramModesClass.getMethod("testMultipleInOutParams", params));
        Object[] paramObjs = new Object[2];
        paramObjs[0] = new Holder<Short>((short)123);
        paramObjs[1] = new Holder<Short>((short)456);
        objContext.setMessageObjects(paramObjs);
        
        corbaContext = new CorbaMessageContextImpl(new GenericMessageContext());
        corbaContext.put(ObjectMessageContext.MESSAGE_INPUT, true);
        
        bindingImpl.marshal(objContext,
                            corbaContext,
                            new JAXBDataBindingCallback(objContext.getMethod(),
                                                        DataBindingCallback.Mode.MESSAGE,
                                                        null));
        
        // Verify that everything was marshalled correctly
        CorbaStreamable[] streamables = corbaContext.getMessage().getStreamableArguments();
        assertTrue(streamables.length == 2);
        CorbaStreamable param0 = streamables[0];
        assertTrue(param0.getObject().getTypeCodeKind().value() == TCKind._tk_short);
        assertTrue(param0.getMode() == ARG_INOUT.value);
        CorbaStreamable param1 = streamables[1];
        assertTrue(param1.getObject().getTypeCodeKind().value() == TCKind._tk_short);
        assertTrue(param1.getMode() == ARG_INOUT.value);

        // Now try the call with a return value
        objContext = new ObjectMessageContextImpl();
        objContext.setMethod(paramModesClass.getMethod("testMultipleInOutParamsWithReturn", params));
        objContext.setMessageObjects(paramObjs);
        
        corbaContext = new CorbaMessageContextImpl(new GenericMessageContext());
        corbaContext.put(ObjectMessageContext.MESSAGE_INPUT, true);
        
        bindingImpl.marshal(objContext,
                            corbaContext,
                            new JAXBDataBindingCallback(objContext.getMethod(),
                                                        DataBindingCallback.Mode.MESSAGE,
                                                        null));
        
        // Verify that everything was marshalled correctly
        streamables = corbaContext.getMessage().getStreamableArguments();
        assertTrue(streamables.length == 2);
        param0 = streamables[0];
        assertTrue(param0.getObject().getTypeCodeKind().value() == TCKind._tk_short);
        assertTrue(param0.getMode() == ARG_INOUT.value);
        param1 = streamables[1];
        assertTrue(param1.getObject().getTypeCodeKind().value() == TCKind._tk_short);
        assertTrue(param1.getMode() == ARG_INOUT.value);
        CorbaStreamable ret = corbaContext.getMessage().getStreamableReturn();
        assertNotNull(ret);
        assertTrue(ret.getObject().getTypeCodeKind().value() == TCKind._tk_short);
    }

    public void testMarshalMultipleMixedParams() throws Exception {
        ObjectMessageContextImpl objContext = null;
        CorbaMessageContextImpl corbaContext = null;
        Class[] params = null;
        
        // First try call the operation using a void return type
        params = new Class[3];
        params[0] = short.class;
        params[1] = new Holder<Short>().getClass();
        params[2] = new Holder<Short>().getClass();
        
        objContext = new ObjectMessageContextImpl();
        objContext.setMethod(paramModesClass.getMethod("testMultipleMixedParams", params));
        Object[] paramObjs = new Object[3];
        paramObjs[0] = (short)123;
        paramObjs[1] = new Holder<Short>((short)456);
        paramObjs[2] = new Holder<Short>((short)789);
        objContext.setMessageObjects(paramObjs);
        
        corbaContext = new CorbaMessageContextImpl(new GenericMessageContext());
        corbaContext.put(ObjectMessageContext.MESSAGE_INPUT, true);
        
        bindingImpl.marshal(objContext,
                            corbaContext,
                            new JAXBDataBindingCallback(objContext.getMethod(),
                                                        DataBindingCallback.Mode.MESSAGE,
                                                        null));
        
        // Verify that everything was marshalled correctly
        CorbaStreamable[] streamables = corbaContext.getMessage().getStreamableArguments();
        assertTrue(streamables.length == 3);
        CorbaStreamable param0 = streamables[0];
        assertTrue(param0.getObject().getTypeCodeKind().value() == TCKind._tk_short);
        assertTrue(param0.getMode() == ARG_IN.value);
        CorbaStreamable param1 = streamables[1];
        assertTrue(param1.getObject().getTypeCodeKind().value() == TCKind._tk_short);
        assertTrue(param1.getMode() == ARG_OUT.value);
        CorbaStreamable param2 = streamables[2];
        assertTrue(param2.getObject().getTypeCodeKind().value() == TCKind._tk_short);
        assertTrue(param2.getMode() == ARG_INOUT.value);

        // Now try the call with a return value
        objContext = new ObjectMessageContextImpl();
        objContext.setMethod(paramModesClass.getMethod("testMultipleMixedParamsWithReturn", params));
        objContext.setMessageObjects(paramObjs);
        
        corbaContext = new CorbaMessageContextImpl(new GenericMessageContext());
        corbaContext.put(ObjectMessageContext.MESSAGE_INPUT, true);
        
        bindingImpl.marshal(objContext,
                            corbaContext,
                            new JAXBDataBindingCallback(objContext.getMethod(),
                                                        DataBindingCallback.Mode.MESSAGE,
                                                        null));
        
        // Verify that everything was marshalled correctly
        streamables = corbaContext.getMessage().getStreamableArguments();
        assertTrue(streamables.length == 3);
        param0 = streamables[0];
        assertTrue(param0.getObject().getTypeCodeKind().value() == TCKind._tk_short);
        assertTrue(param0.getMode() == ARG_IN.value);
        param1 = streamables[1];
        assertTrue(param1.getObject().getTypeCodeKind().value() == TCKind._tk_short);
        assertTrue(param1.getMode() == ARG_OUT.value);
        param2 = streamables[2];
        assertTrue(param2.getObject().getTypeCodeKind().value() == TCKind._tk_short);
        assertTrue(param2.getMode() == ARG_INOUT.value);
        CorbaStreamable ret = corbaContext.getMessage().getStreamableReturn();
        assertNotNull(ret);
        assertTrue(ret.getObject().getTypeCodeKind().value() == TCKind._tk_short);
    }
    
    public void testMarshalReturn() throws Exception {
        ObjectMessageContextImpl objContext = null;
        CorbaMessageContextImpl corbaContext = null;
        Class[] params = null;
        
        objContext = new ObjectMessageContextImpl();
        objContext.setMethod(paramModesClass.getMethod("testReturn", params));
        
        corbaContext = new CorbaMessageContextImpl(new GenericMessageContext());
        corbaContext.put(ObjectMessageContext.MESSAGE_INPUT, true);
        
        bindingImpl.marshal(objContext,
                            corbaContext,
                            new JAXBDataBindingCallback(objContext.getMethod(),
                                                        DataBindingCallback.Mode.MESSAGE,
                                                        null));
        
        CorbaStreamable ret = corbaContext.getMessage().getStreamableReturn();
        assertNotNull(ret);
        assertTrue(ret.getObject().getTypeCodeKind().value() == TCKind._tk_short);
    }
    
    
}
