/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */
package org.apache.yoko.bindings.corba;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.math.BigDecimal;

import javax.xml.namespace.QName;
import javax.xml.stream.XMLEventFactory;
import javax.xml.stream.XMLEventReader;
import javax.xml.stream.XMLEventWriter;
import javax.xml.stream.XMLInputFactory;
import javax.xml.stream.XMLOutputFactory;
import javax.xml.stream.XMLStreamException;

import junit.framework.TestCase;

import org.apache.schemas.yoko.bindings.corba.Enum;
import org.apache.schemas.yoko.bindings.corba.Fixed;
import org.apache.schemas.yoko.bindings.corba.Struct;
import org.apache.yoko.bindings.corba.types.CorbaEnumHandler;
import org.apache.yoko.bindings.corba.types.CorbaFixedHandler;
import org.apache.yoko.bindings.corba.types.CorbaObjectHandler;
import org.apache.yoko.bindings.corba.types.CorbaPrimitiveHandler;
import org.apache.yoko.bindings.corba.types.CorbaStructHandler;
import org.apache.yoko.wsdl.CorbaConstants;
import org.apache.yoko.wsdl.CorbaTypeImpl;
import org.apache.yoko.orb.CORBA.OutputStream;
import org.apache.yoko.orb.OCI.Buffer;

import org.objectweb.celtix.Bus;
import org.objectweb.celtix.bindings.BindingManager;
import org.objectweb.celtix.ws.addressing.EndpointReferenceType;

import org.omg.CORBA.ORB;
import org.omg.CORBA.portable.InputStream;
import org.omg.CORBA.TypeCode;

public class CorbaNestedTypeTest extends TestCase {

    private final String nestedNamespaceURI = "http://yoko.apache.org/nested/idl_types";
    private final String nestedPrefix = "corbatm";
    
    private TestUtils testUtils;
    private ORB orb;
    private Bus bus;
    private EndpointReferenceType endpoint;
    private CorbaBindingImpl bindingImpl;
    
    public CorbaNestedTypeTest(String arg0) {
        super(arg0);
    }
    public static void main(String[] args) {
        junit.textui.TestRunner.run(CorbaBindingImplMarshalTest.class);
    }
    
    protected void setUp() throws Exception {
        super.setUp();
    
        testUtils = new TestUtils();
    
        java.util.Properties props = System.getProperties();
        props.put("org.omg.CORBA.ORBClass", "org.apache.yoko.orb.CORBA.ORB");
        props.put("org.omg.CORBA.ORBSingletonClass", "org.apache.yoko.orb.CORBA.ORBSingleton");
        props.put("yoko.orb.id", "Yoko-Test");
        orb = ORB.init(new String[0], props);
        bus = Bus.init(new String[0]);
        
        // programmatically add the corba binding factory
        BindingManager manager = bus.getBindingManager();
        CorbaBindingFactory corbaBF = new CorbaBindingFactory();
        corbaBF.init(bus);
        manager.registerBinding(CorbaConstants.NU_WSDL_CORBA, corbaBF);
        
        endpoint = testUtils.getNestedEndpointReference();
        bindingImpl = new CorbaBindingImpl(bus, endpoint, orb, false);
    }
    
    protected void tearDown() throws Exception {
        if (orb != null) {
            try {
                orb.destroy();
            } catch (Exception ex) {
                // Do nothing.  Throw an Exception?
            }
        }
    }
    
    // Since the nested type aspect of this Binding really just involves the CorbaObjectHandlers and the 
    // reading an writing to and from Stax, we'll test just these components.
    
    public void testNestedType() {
        QName objName = new QName("object");
        QName objIdlType = new QName(nestedNamespaceURI, "NestedStruct", nestedPrefix);
        CorbaTypeImpl type = CorbaUtils.getCorbaType(objIdlType, bindingImpl.getCorbaTypeMaps());
        assertTrue(type instanceof Struct);
        Struct structType = (Struct)type;
        TypeCode objTypeCode = CorbaUtils.getTypeCode(orb, objIdlType, bindingImpl.getCorbaTypeMaps());
        CorbaStructHandler obj = new CorbaStructHandler(objName, objIdlType, objTypeCode, structType);
        assertNotNull(obj);
        
        // Create member 1, a BasicStruct
        objName = new QName("member1");
        objIdlType = new QName(nestedNamespaceURI, "BasicStruct", nestedPrefix);
        type = CorbaUtils.getCorbaType(objIdlType, bindingImpl.getCorbaTypeMaps());
        assertTrue(type instanceof Struct);
        structType = (Struct)type;
        objTypeCode = CorbaUtils.getTypeCode(orb, objIdlType, bindingImpl.getCorbaTypeMaps());
        CorbaStructHandler member1Obj = new CorbaStructHandler(objName, objIdlType, objTypeCode, structType);
        assertNotNull(member1Obj);
        
        // Create the basic struct members
        String nestedMember1Value = "Basic Member 1";
        objName = new QName("member1");
        objIdlType = CorbaConstants.NT_CORBA_STRING;
        objTypeCode = CorbaUtils.getTypeCode(orb, objIdlType, bindingImpl.getCorbaTypeMaps());
        CorbaPrimitiveHandler nestedMember1Obj = 
            new CorbaPrimitiveHandler(objName, objIdlType, objTypeCode, null);
        assertNotNull(nestedMember1Obj);
        member1Obj.addMember(nestedMember1Obj);

        Boolean nestedMember2Value = new Boolean(false);
        objName = new QName("member2");
        objIdlType = CorbaConstants.NT_CORBA_BOOLEAN;
        objTypeCode = CorbaUtils.getTypeCode(orb, objIdlType, bindingImpl.getCorbaTypeMaps());
        CorbaPrimitiveHandler nestedMember2Obj = 
            new CorbaPrimitiveHandler(objName, objIdlType, objTypeCode, null);
        assertNotNull(nestedMember2Obj);
        member1Obj.addMember(nestedMember2Obj);
        
        obj.addMember(member1Obj);
        
        // Create member 2, a Basic Fixed
        BigDecimal fixedValue = new BigDecimal("123.67");
        objName = new QName("member2");
        objIdlType = new QName(nestedNamespaceURI, "BasicFixed", nestedPrefix);
        type = CorbaUtils.getCorbaType(objIdlType, bindingImpl.getCorbaTypeMaps());
        assertTrue(type instanceof Fixed);
        Fixed fixedType = (Fixed)type;
        objTypeCode = CorbaUtils.getTypeCode(orb, objIdlType, bindingImpl.getCorbaTypeMaps());
        CorbaFixedHandler member2Obj = new CorbaFixedHandler(objName, objIdlType, objTypeCode, fixedType);
        assertNotNull(member2Obj);
        obj.addMember(member2Obj);
        
        // Create member 3, a Basic Enum
        int enumIndex = 0;
        objName = new QName("member3");
        objIdlType = new QName(nestedNamespaceURI, "BasicEnum", nestedPrefix);
        type = CorbaUtils.getCorbaType(objIdlType, bindingImpl.getCorbaTypeMaps());
        assertTrue(type instanceof Enum);
        Enum enumType = (Enum)type;
        objTypeCode = CorbaUtils.getTypeCode(orb, objIdlType, bindingImpl.getCorbaTypeMaps());
        CorbaEnumHandler member3Obj = new CorbaEnumHandler(objName, objIdlType, objTypeCode, enumType);
        assertNotNull(member3Obj);
        obj.addMember(member3Obj);
        
        // Create member 4, a Basic Enum
        Float floatValue = 123456.678f;
        objName = new QName("member4");
        objIdlType = CorbaConstants.NT_CORBA_FLOAT;
        objTypeCode = CorbaUtils.getTypeCode(orb, objIdlType, bindingImpl.getCorbaTypeMaps());
        CorbaPrimitiveHandler member4Obj = new CorbaPrimitiveHandler(objName, objIdlType, objTypeCode, null);
        assertNotNull(member4Obj);
        obj.addMember(member4Obj);

        // Write to an output stream the way a CORBA application would marshal the data for the nested
        // struct.
        Buffer buf = new Buffer();
        OutputStream oStream = new OutputStream(buf);
        
        oStream.write_string(nestedMember1Value);
        oStream.write_boolean(nestedMember2Value.booleanValue());
        oStream.write_fixed(fixedValue.movePointRight(2));
        oStream.write_long(enumIndex);
        oStream.write_float(floatValue.floatValue());
        
        InputStream iStream = oStream.create_input_stream();
        CorbaObjectReader reader = new CorbaObjectReader(iStream);
        reader.read(obj);        
        
        // Now that the Object has its values, try and write it to Stax
        XMLInputFactory inputFactory = XMLInputFactory.newInstance();
        XMLOutputFactory outputFactory = XMLOutputFactory.newInstance();
        outputFactory.setProperty(XMLOutputFactory.IS_REPAIRING_NAMESPACES, Boolean.TRUE);
      
        ByteArrayOutputStream outStream = new ByteArrayOutputStream();
        XMLEventWriter evtWriter = null;
        try {
            evtWriter = outputFactory.createXMLEventWriter(outStream);
        } catch (XMLStreamException ex) {
            assertTrue(false);
        }
        assertNotNull(evtWriter);
        
        XMLEventFactory factory = XMLEventFactory.newInstance();
        try {
            bindingImpl.writeObjectToStax(obj, evtWriter, factory, false);
            evtWriter.flush();
        } catch (XMLStreamException ex) {
            assertTrue(false);
        }
        
        byte[] b = outStream.toByteArray();
        for (int i = 0; i < b.length; ++i) {
            System.out.print((char)b[i]);
        }
        System.out.println();
        
        ByteArrayInputStream inStream = new ByteArrayInputStream(outStream.toByteArray());        
        XMLEventReader evtReader = null;
        try {            
            evtReader = inputFactory.createXMLEventReader(inStream);
        } catch (XMLStreamException ex) {
            assertTrue(false);
        }
        assertNotNull(evtReader);
        
        // Now use the Stax data to re-create a new object
        CorbaObjectHandler result = 
            bindingImpl.readObjectFromStax(evtReader, 
                                           new QName(nestedNamespaceURI, "NestedStruct", nestedPrefix), 
                                           false);
        
        // Verify that the object has the correct values
        assertTrue(result instanceof CorbaStructHandler);        
        CorbaStructHandler nestedStructResult = (CorbaStructHandler)result;
        
        result = nestedStructResult.getMember(0);
        assertTrue(result instanceof CorbaStructHandler);
        CorbaStructHandler structResult = (CorbaStructHandler)result;

        // These are from the struct nested inside our other struct
        result = structResult.getMember(0);
        assertTrue(result instanceof CorbaPrimitiveHandler);
        CorbaPrimitiveHandler structPrimitiveMemberResult = (CorbaPrimitiveHandler)result;
        assertTrue(structPrimitiveMemberResult.getValueData().equals(nestedMember1Value.toString()));
        result = structResult.getMember(1);
        assertTrue(result instanceof CorbaPrimitiveHandler);
        structPrimitiveMemberResult = (CorbaPrimitiveHandler)result;
        assertTrue(structPrimitiveMemberResult.getValueData().equals(nestedMember2Value.toString()));
        
        // This is the fixed nested inside our struct
        result = nestedStructResult.getMember(1);
        assertTrue(result instanceof CorbaFixedHandler);
        CorbaFixedHandler fixedResult = (CorbaFixedHandler)result;
        assertTrue(fixedResult.getValue().equals(fixedValue));

        // This is the enum nested inside our struct
        result = nestedStructResult.getMember(2);
        assertTrue(result instanceof CorbaEnumHandler);
        CorbaEnumHandler enumResult = (CorbaEnumHandler)result;
        assertTrue(enumResult.getValue().equals(enumType.getEnumerator().get(enumIndex).getValue()));
        
        // This is the primitive nested inside our struct
        result = nestedStructResult.getMember(3);
        assertTrue(result instanceof CorbaPrimitiveHandler);
        CorbaPrimitiveHandler primitiveResult = (CorbaPrimitiveHandler)result;
        assertTrue(primitiveResult.getValueData().equals(floatValue.toString()));
    }
    
}
