/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements. See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership. The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
*/

package org.apache.yoko.tools.processors.idl;

import javax.wsdl.Definition;
import javax.wsdl.Message;

import javax.xml.namespace.QName;

import antlr.ASTVisitor;
import antlr.collections.AST;

import org.apache.schemas.yoko.bindings.corba.ArgType;
import org.apache.schemas.yoko.bindings.corba.ModeType;
import org.apache.schemas.yoko.bindings.corba.OperationType;
import org.apache.schemas.yoko.bindings.corba.ParamType;

import org.apache.ws.commons.schema.XmlSchemaElement;
import org.apache.ws.commons.schema.XmlSchemaSequence;

public class MessagePartVisitor implements ASTVisitor {

    static final String IN_PARAMETER = "inparameter";
    static final String OUT_PARAMETER = "outparameter";
    static final String INOUT_PARAMETER = "inoutparameter";
    static final String RETURN_PARAMETER = "return";

    Definition definition;
    Message inputMsg;
    Message outputMsg;

    XmlSchemaSequence outWrappingSequence;
    XmlSchemaElement outWrappedElement;
    XmlSchemaSequence inWrappingSequence;
    XmlSchemaElement inWrappedElement;

    OperationType corbaOperation;

    TypesVisitor typesVisitor;

    public MessagePartVisitor(WSDLASTVisitor visitor,
                              XmlSchemaSequence inWrapSeq,
                              XmlSchemaSequence outWrapSeq) {
        definition = visitor.getDefinition();   
        typesVisitor = visitor.getTypesVisitor();
        inWrappingSequence = inWrapSeq;
        outWrappingSequence = outWrapSeq;
    }

    public void setInputMessage(Message msg) {
        inputMsg = msg;
    }

    public void setOutputMessage(Message msg) {
        outputMsg = msg;
    }

    public Message getInputMessage() {
        return inputMsg;
    }

    public Message getOutputMessage() {
        return outputMsg;
    }

    public void setCorbaOperation(OperationType op) {
        corbaOperation = op;
    }
    
    //Assumption for now is to generate wrapped doc-literal wsdl.

    public void visit(AST node) {
        switch (node.getType()) {
        case IDLTokenTypes.LITERAL_in: {
            AST typeNode = node.getFirstChild();
            String partName = TypesUtils.getPrimitiveCorbaTypeNameNode(typeNode).toString();
            ArgType param = createCorbaParam(partName, ModeType.IN);
            createInputParameter(partName, param, typeNode);
            break;
        }
        case IDLTokenTypes.LITERAL_inout: {
            AST typeNode = node.getFirstChild();
            String partName = TypesUtils.getPrimitiveCorbaTypeNameNode(typeNode).toString();
            ArgType param = createCorbaParam(partName, ModeType.INOUT);
            createInputParameter(partName, param, typeNode);
            createOutputParameter(partName, param, typeNode);
            break;
        }
        case IDLTokenTypes.LITERAL_out: {
            AST typeNode = node.getFirstChild();
            String partName = TypesUtils.getPrimitiveCorbaTypeNameNode(typeNode).toString();
            ArgType param = createCorbaParam(partName, ModeType.OUT);
            createOutputParameter(partName, param, typeNode);
            break;
        }
        case IDLTokenTypes.LITERAL_void: {
            // nothing to do here
            break;
        }
        default: {
            String partName = RETURN_PARAMETER;
            ArgType param = createCorbaReturn(partName);
            createOutputParameter(partName, param, node);
        }
        }
    }

    private void createInputParameter(String partName, ArgType param, AST typeNode) {
        QName element = inputMsg.getQName();
        inWrappedElement = typesVisitor.addElement(inWrappingSequence, element, partName);
        typesVisitor.setCurrentPart(inWrappedElement, partName, param);
        typesVisitor.visit(typeNode);
    }

    private void createOutputParameter(String partName, ArgType param, AST typeNode) {
        QName element = outputMsg.getQName();
        outWrappedElement = typesVisitor.addElement(outWrappingSequence, element, partName);
        typesVisitor.setCurrentPart(outWrappedElement, partName, param);
        typesVisitor.visit(typeNode);
    }

    private ArgType createCorbaParam(String partName, ModeType mode) {
        ParamType param = new ParamType();
        param.setName(partName);
        param.setMode(mode);
        corbaOperation.getParam().add(param);
        return param;
    }

    private ArgType createCorbaReturn(String partName) {
        ArgType param = new ArgType();
        param.setName(partName);
        corbaOperation.setReturn(param);
        return param;
    }
}
