/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements. See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership. The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
*/

package org.apache.yoko.tools.processors.idl;

import javax.xml.namespace.QName;

import antlr.collections.AST;

import org.apache.schemas.yoko.bindings.corba.Alias;
import org.apache.schemas.yoko.bindings.corba.TypeMappingType;

import org.apache.ws.commons.schema.XmlSchema;
import org.apache.ws.commons.schema.XmlSchemaCollection;
import org.apache.ws.commons.schema.XmlSchemaSimpleType;
import org.apache.ws.commons.schema.XmlSchemaSimpleTypeRestriction;
import org.apache.ws.commons.schema.XmlSchemaType;
import org.apache.yoko.wsdl.CorbaConstants;

public class TypedefVisitor extends TypesVisitorBase {

    private static final int SEQUENCE = -1;
    private static final int FIXED =    -2;
    private static final int STRUCT =   -3;
    private static final int ENUM =     -4;
    private static final int UNION =    -5;
    
    public TypedefVisitor(XmlSchemaCollection xmlSchemas,
                          XmlSchema xmlSchema,
                          TypeMappingType typeMappingType) {
        super(xmlSchemas, xmlSchema, typeMappingType);
    }
    
    public void visit(AST typedefNode) {
        // "typedef" <type_declarator>
        // <type_declarator> ::= <type_spec> <declarators>
        // <type_spec> ::= <simple_type_spec>
        //               | <constr_type_spec>
        // <simple_type_spec> ::= <base_type_spec>
        //                      | <template_type_spec> 
        //                      | <scoped_name>
        // <base_type_spec> ::= ... omitted (integer, char, octect, etc)
        // <template_type_spec> ::= <sequence_type>
        //                        | <string_type>
        //                        | <wstring_type>
        //                        | <fixed_pt_type>
        // <constr_type_spec> ::= <struct_type>
        //                      | <union_type>
        //                      | <enum_type>
        // <declarators> ::= <declarator> {"," <declarator>}*
        // <declarator> ::= <simple_declarator>
        //                | <complex_declarator>
        // <simple_declarator> ::= <identifier>
        // <complex_declarator> ::= <array_declarator>
        // <array_declarator> ::= <identifier> <fixed_array_size>+
        // <fixed_array_size> ::= "[" <positive_int_const> "]"

        //AST typeDeclaratorNode = typedefNode.getFirstChild();

        TypesVisitorBase visitor = null;
        int type = getType(typedefNode);
        switch (type) {
        case SEQUENCE: {
            visitor = new ArrayVisitor(schemas, schema, typeMap);
            visitor.visit(typedefNode);
            break;
        }
        case FIXED: {
            visitor = new FixedVisitor(schemas, schema, typeMap);
            visitor.visit(typedefNode);
            break;
        }
        case STRUCT: {
            visitor = new StructVisitor(schemas, schema, typeMap);
            AST structNode = typedefNode.getFirstChild(); 
            visitor.visit(structNode);
            visitAlias(typedefNode);
            break;
        }
        case ENUM: {
            visitor = new EnumVisitor(schemas, schema, typeMap);
            visitor.visit(typedefNode.getFirstChild());
            visitAlias(typedefNode);
            break;
        }
        case UNION: {
            visitor = new UnionVisitor(schemas, schema, typeMap);
            visitor.visit(typedefNode.getFirstChild());
            visitAlias(typedefNode);
            break;    
        }
        default:
            visitAlias(typedefNode);
            break;
        }
    }
    
    private int getType(AST node) {
        if (isSequence(node)) {
            return SEQUENCE;
        } else if (isFixed(node)) {
            return FIXED;
        } else if (isStruct(node)) {
            return STRUCT;
        } else if (isEnum(node)) {
            return ENUM;
        } else if (isUnion(node)) {
            return UNION;
        } else {
            return node.getType();
        }
    }

    private boolean isSequence(AST node) {
        boolean result = false;
        if (node.getType() == IDLTokenTypes.LITERAL_typedef) {
            AST node2 = node.getFirstChild();
            if (node2.getType() == IDLTokenTypes.LITERAL_sequence) {                
                result = true;
            }
        }
        return result;
    }

    private boolean isFixed(AST node) {
        boolean result = false;
        if (node.getType() == IDLTokenTypes.LITERAL_typedef) {
            AST node2 = node.getFirstChild();
            if (node2.getType() == IDLTokenTypes.LITERAL_fixed) {                
                result = true;
            }
        }
        return result;
    }

    private boolean isStruct(AST node) {
        boolean result = false;
        if (node.getType() == IDLTokenTypes.LITERAL_typedef) {
            AST node2 = node.getFirstChild();
            if (node2.getType() == IDLTokenTypes.LITERAL_struct) {                
                result = true;
            }
        }
        return result;        
    }
    
    private boolean isEnum(AST node) {
        boolean result = false;
        if (node.getType() == IDLTokenTypes.LITERAL_typedef) {
            AST node2 = node.getFirstChild();
            if (node2.getType() == IDLTokenTypes.LITERAL_enum) {                
                result = true;
            }
        }
        return result;                
    }
    
    private boolean isUnion(AST node) {
        boolean result = false;
        if (node.getType() == IDLTokenTypes.LITERAL_typedef) {
            AST node2 = node.getFirstChild();
            if (node2.getType() == IDLTokenTypes.LITERAL_union) {                
                result = true;
            }
        }
        return result;                
    }
    
    private void visitAlias(AST typedefNode) {
        AST typeSpecNode = typedefNode.getFirstChild();
        
        // perhaps getPrimitiveCorbaTypeNameNode is a misnomer
        AST typeNameNode = TypesUtils.getPrimitiveCorbaTypeNameNode(typeSpecNode); 

        XmlSchemaType type = null;
        switch (typeSpecNode.getType()) {
        case IDLTokenTypes.LITERAL_struct:
        case IDLTokenTypes.LITERAL_enum:
        case IDLTokenTypes.LITERAL_union:
            type = TypesUtils.findType(schemas, schema, typeSpecNode.getFirstChild());
            break;
        default:
            type = TypesUtils.findType(schemas, schema, typeSpecNode);
            break;
        }
        
        // xmlschema:simpleType
        XmlSchemaSimpleType simpleType = new XmlSchemaSimpleType(schema);
        simpleType.setName(typeNameNode.toString());
        XmlSchemaSimpleTypeRestriction restriction = new XmlSchemaSimpleTypeRestriction();
        restriction.setBaseTypeName(type.getQName());
        simpleType.setContent(restriction);
        
        // add xmlschema:simpleType
        getSchema().getItems().add(simpleType);
        getSchema().addType(simpleType);
        
        
        // corba:alias
        Alias alias = new Alias();
        alias.setQName(new QName(typeMap.getTargetNamespace(), typeNameNode.toString()));
        alias.setBasetype(TypesUtils.findCorbaType(typeMap, type.getQName()).getQName());
        alias.setType(type.getQName());
        alias.setRepositoryID(CorbaConstants.REPO_STRING
                              + typeNameNode.toString()
                              + CorbaConstants.IDL_VERSION);
        
        // add corba:alias
        typeMap.getStructOrExceptionOrUnion().add(alias);
    }
}
